local MAJOR = "LibScriptableDisplayRGBA-1.0" 
local MINOR = 15

assert(LibStub, MAJOR.." requires LibStub") 
local LibRGBA = LibStub:NewLibrary(MAJOR, MINOR)
if not LibRGBA then return end

local pool = setmetatable({}, {__mode = "k"})

if not LibRGBA.__index then
	LibRGBA.__index = LibRGBA
end

-- Create a new LibScriptableDisplayRGBAText object
-- @name WidgetText.New
-- @usage WidgetText:New(visitor, name, config, row, col, layer, type, environment, errorLevel)
-- @param this The inheritance object
-- @param visitor An LibScriptableDisplayCore-1.0 object
-- @parma name This widget's name
-- @param config This widget's configuration
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param type Dict of widget types.
-- @param environment The code execution environment.
-- @param errorLevel The error level for this object.
-- @return A new LibScriptableDisplayRGBAText object
function LibRGBA:New(r, g, b, a) 
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end
					
	setmetatable(obj, self)

	obj.r = r or 0xff
	obj.g = g or 0xff
	obj.b = b or 0xff
	obj.a = a or 0xff

	return obj	
end


-- @name LibScriptableDisplayRGBA:Del
-- @usage LibScriptableDisplayRGBA:Del([widget]) or object:Del()
-- @param widget An optional widget object
-- @return Nothing
function LibRGBA:Del(widget)
	if type(widget) ~= "table" then widget = self end
	pool[widget] = true
end

-- @name LibScriptableDisplayRGBA.Compare
-- @usage STarLibRGBA:Compare(other)
-- @param other The other LibScriptableDisplayRGBA object to compare self to
-- @return True if the two colors (self and other) compare equally
function LibRGBA:Compare(other)
	local equal = true
	if self.R ~= other.R or other.G ~= v.G or self.B ~= other.B or self.A ~= other.A then
		equal = false
	end
	return equal
end

